@@ -32,6 +32,7 @@ name             # of qubits args                                   description
 measurement      >= 1        ``key: str``                           Measurement in the Z basis.
 phased_rx        1           ``angle_t: float``, ``phase_t: float`` Phased x-rotation gate.
 cz               2                                                  Controlled-Z gate.
+barrier          >= 2                                               Barrier instruction.
 ================ =========== ====================================== ===========
 
 Measurement
@@ -76,16 +77,29 @@ Symmetric wrt. the qubits it's acting on. Takes no arguments.
 Example: ``Instruction(name='cz', qubits=['alice', 'bob'], args={})``
 
 
+Barrier
+-------
+
+Barriers ensure that all operations after the barrier on the qubit subsystems spanned by
+the barrier are only executed when all the operations before the barrier have been completed.
+
+Example: ``Instruction(name='barrier', qubits=['alice', 'bob'], args={})``
+
+
 Circuit output
 ==============
 
-The :class:`RunResult` class represents the results of a quantum circuit execution.
-If the run succeeded, :attr:`RunResult.measurements` contains the output of the circuit, consisting
-of the results of the measurement operations in the circuit.
-It is a dictionary that maps each measurement key to a 2D array of measurement results, represented as a nested list.
-``RunResult.measurements[key][shot][index]`` is the result of measuring the ``index``'th qubit in measurement
-operation ``key`` in the shot ``shot``. The results are nonnegative integers representing the computational
-basis state (for qubits, 0 or 1) that was the measurement outcome.
+The :class:`RunResult` class represents the results of the quantum circuit execution.
+If the run succeeded, :attr:`RunResult.measurements` contains the output of the batch of circuits,
+consisting of the results of the measurement operations in each circuit.
+It is a list of dictionaries, where each dict maps each measurement key to a 2D array of measurement
+results, represented as a nested list.
+``RunResult.measurements[circuit_index][key][shot][qubit_index]`` is the result of measuring the
+``qubit_index``'th qubit in measurement operation ``key`` in the shot ``shot`` in the
+``circuit_index``'th circuit of the batch.
+
+The results are nonnegative integers representing the computational basis state (for qubits, 0 or 1)
+that was the measurement outcome.
 
 ----
 """
@@ -105,6 +119,8 @@ from uuid import UUID
 import requests
 from pydantic import BaseModel, Field
 
+REQUESTS_TIMEOUT = 10
+
 DEFAULT_TIMEOUT_SECONDS = 900
 SECONDS_BETWEEN_CALLS = 1
 REFRESH_MARGIN_SECONDS = 5
@@ -133,7 +149,7 @@ class APITimeoutError(CircuitExecutionError):
     """
 
 
-class RunStatus(str, Enum):
+class Status(str, Enum):
     """
     Status of a task.
     """
@@ -169,6 +185,14 @@ class Circuit(BaseModel):
     instructions: list[Instruction] = Field(..., description='instructions comprising the circuit')
     'instructions comprising the circuit'
 
+    def all_qubits(self) -> set[str]:
+        """Return the names of all qubits in the circuit.
+        """
+        qubits = set()
+        for instruction in self.instructions:
+            qubits.update(instruction.qubits)
+        return qubits
+
 
 class SingleQubitMapping(BaseModel):
     """Mapping of a logical qubit name to a physical qubit name.
@@ -179,38 +203,85 @@ class SingleQubitMapping(BaseModel):
     'physical qubit name'
 
 
-class RunRequest(BaseModel):
-    """Request for an IQM quantum computer to execute a quantum circuit.
+def serialize_qubit_mapping(qubit_mapping: dict[str, str]) -> list[SingleQubitMapping]:
+    """Serializes a qubit mapping dict into the corresponding IQM data transfer format.
+
+    Args:
+        qubit_mapping: mapping from logical to physical qubit names
+
+    Returns:
+        data transfer object representing the mapping
     """
-    circuit: Circuit = Field(..., description='quantum circuit to execute')
-    'quantum circuit to execute'
-    settings: dict[str, Any] = Field(..., description='EXA settings node containing the calibration data')
-    'EXA settings node containing the calibration data'
+    return [SingleQubitMapping(logical_name=k, physical_name=v) for k, v in qubit_mapping.items()]
+
+
+class RunRequest(BaseModel):
+    """Request for an IQM quantum computer to execute a batch of quantum circuits.
+
+    Note: all circuits in a batch must measure the same qubits otherwise batch execution fails.
+    """
+    circuits: list[Circuit] = Field(..., description='batch of quantum circuit(s) to execute')
+    'batch of quantum circuit(s) to execute'
+    settings: Optional[dict[str, Any]] = Field(
+        None,
+        description='IQM hardware settings and calibration data, or None to use the latest calibration set'
+    )
+    'IQM hardware settings and calibration data, or None to use the latest calibration set'
+    calibration_set_id: Optional[int] = Field(
+        None,
+        description='ID of the calibration set to use, or None to use the latest calibration set'
+    )
+    'ID of the calibration set to use, or None to use the latest calibration set'
     qubit_mapping: Optional[list[SingleQubitMapping]] = Field(
         None,
-        description='mapping of logical qubit names to physical qubit names'
+        description='mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
+    )
+    'mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
+    shots: int = Field(..., description='how many times to execute each circuit in the batch')
+    'how many times to execute each circuit in the batch'
+
+
+CircuitMeasurementResults = dict[str, list[list[int]]]
+"""Measurement results from a single circuit. For each measurement operation in the circuit,
+maps the measurement key to the corresponding results. The outer list elements correspond to shots,
+and the inner list elements to the qubits measured in the measurement operation."""
+
+
+class Metadata(BaseModel):
+    """Metadata belonging to a job sumission"""
+    shots: int = Field(..., description='how many times to execute each circuit in the batch')
+    'how many times to execute each circuit in the batch'
+    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(
+        None,
+        description='mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
+    )
+    'mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
+    circuits: list[Circuit] = Field(..., description='batch of quantum circuit(s) to execute')
+    'batch of quantum circuit(s) to execute'
+    calibration_set_id: Optional[int] = Field(
+        None,
+        description='ID of the calibration set used, or None if settings were specified'
     )
-    'mapping of logical qubit names to physical qubit names'
-    shots: int = Field(..., description='how many times to execute the circuit')
-    'how many times to execute the circuit'
 
 
 class RunResult(BaseModel):
-    """Results of a circuit execution.
+    """Results of executing a batch of circuit(s).
 
     * ``measurements`` is present iff the status is ``'ready'``.
     * ``message`` carries additional information for the ``'failed'`` status.
     * If the status is ``'pending'``, ``measurements`` and ``message`` are ``None``.
     """
-    status: RunStatus = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
+    status: Status = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
     "current status of the run, in ``{'pending', 'ready', 'failed'}``"
-    measurements: Optional[dict[str, list[list[int]]]] = Field(
+    measurements: Optional[list[CircuitMeasurementResults]] = Field(
         None,
-        description='if the run has finished successfully, the measurement results for the circuit'
+        description='if the run has finished successfully, the measurement results for the circuit(s)'
     )
-    'if the run has finished successfully, the measurement results for the circuit'
+    'if the run has finished successfully, the measurement results for the circuit(s)'
     message: Optional[str] = Field(None, description='if the run failed, an error message')
     'if the run failed, an error message'
+    metadata: Metadata = Field(..., description='metadata about the underlying job request')
+    'metadata about the underlying job request'
     warnings: Optional[list[str]] = Field(None, description='list of warning messages')
     'list of warning messages'
 
@@ -226,7 +297,31 @@ class RunResult(BaseModel):
 
         """
         input_copy = inp.copy()
-        return RunResult(status=RunStatus(input_copy.pop('status')), **input_copy)
+        return RunResult(status=Status(input_copy.pop('status')), **input_copy)
+
+
+class RunStatus(BaseModel):
+    """Status of a batchcircuit execution request."""
+    status: Status = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
+    "current status of the run, in ``{'pending', 'ready', 'failed'}``"
+    message: Optional[str] = Field(None, description='if the run failed, an error message')
+    'if the run failed, an error message'
+    warnings: Optional[list[str]] = Field(None, description='list of warning messages')
+    'list of warning messages'
+
+    @staticmethod
+    def from_dict(inp: dict[str, Union[str, dict]]) -> RunStatus:
+        """Parses the result from a dict.
+
+        Args:
+            inp: value to parse, has to map to RunResult
+
+        Returns:
+            parsed run result
+
+        """
+        input_copy = inp.copy()
+        return RunStatus(status=Status(input_copy.pop('status')), **input_copy)
 
 
 class GrantType(str, Enum):
@@ -267,7 +362,7 @@ class AuthRequest(BaseModel):
 class Credentials(BaseModel):
     """Credentials and tokens for maintaining a session with the authentication server.
 
-    * Fields ``auth:server_url``, ``username`` and ``password`` are provided by the user.
+    * Fields ``auth_server_url``, ``username`` and ``password`` are provided by the user.
     * Fields ``access_token`` and ``refresh_token`` are loaded from the authentication server and
       refreshed periodically.
     """
@@ -283,6 +378,19 @@ class Credentials(BaseModel):
     'current refresh token of the session'
 
 
+class ExternalToken(BaseModel):
+    """Externally managed token for maintaining a session with the authentication server.
+
+    * Fields ``auth_server_url`` and ``access_token`` are loaded from an
+      external resource, e.g. file generated by Cortex CLI's token manager.
+    """
+
+    auth_server_url: str = Field(..., description='Base URL of the authentication server')
+    'Base URL of the authentication server'
+    access_token: str = Field(None, description='current access token of the session')
+    'current access token of the session'
+
+
 def _get_credentials(credentials: dict[str, str]) -> Optional[Credentials]:
     """Try to obtain credentials, first from arguments, then from environment variables.
 
@@ -302,6 +410,40 @@ def _get_credentials(credentials: dict[str, str]) -> Optional[Credentials]:
     return Credentials(auth_server_url=auth_server_url, username=username, password=password)
 
 
+def _get_external_token(tokens_file: Optional[str] = None) -> Optional[ExternalToken]:
+    """Try to obtain external token from a file, first by path provided, then by path from
+    environment variable.
+
+    Args:
+        tokens_file: path to a JSON file containing tokens
+
+    Returns:
+        ExternalToken with non-empty auth_server_url and access_token fields,
+        or None if ``tokens_file`` was not provided.
+    """
+
+    filepath = tokens_file or os.environ.get('IQM_TOKENS_FILE')
+
+    if not filepath:
+        return None
+
+    try:
+        with open(filepath, 'r', encoding='utf-8') as file:
+            raw_data = file.read()
+    except FileNotFoundError as error:
+        raise ClientConfigurationError(f'File not found: {filepath}') from error
+
+    try:
+        json_data = json.loads(raw_data)
+    except json.decoder.JSONDecodeError as error:
+        raise ClientConfigurationError(f'Decoding JSON has failed, {error}') from error
+
+    auth_server_url = json_data['auth_server_url']
+    access_token = json_data['access_token']
+
+    return ExternalToken(auth_server_url=auth_server_url, access_token=access_token)
+
+
 def _time_left_seconds(token: str) -> int:
     """Check how much time is left until the token expires.
 
@@ -320,7 +462,10 @@ class IQMClient:
 
     Args:
         url: Endpoint for accessing the server. Has to start with http or https.
-        settings: Settings for the quantum computer, in IQM JSON format.
+        tokens_file: Optional path to a tokens file used for authentication.
+            This can also be set in the IQM_TOKENS_FILE environment variable.
+            If tokens_file is set, auth_server_url, username and password
+            must not be set.
 
     Keyword Args:
         auth_server_url: Optional base URL of the authentication server.
@@ -333,43 +478,87 @@ class IQMClient:
             This can also be set in the IQM_AUTH_PASSWORD environment variable.
             Password must be set if ``auth_server_url`` is set.
     """
+
     def __init__(
             self,
             url: str,
-            settings: dict[str, Any],
+            tokens_file: Optional[str] = None,
             **credentials  # contains auth_server_url, username, password
     ):
         if not url.startswith(('http:', 'https:')):
             raise ClientConfigurationError(f'The URL schema has to be http or https. Incorrect schema in URL: {url}')
+        if tokens_file and credentials:
+            raise ClientConfigurationError('Either external token or credentials must be provided. Both were provided.')
         self._base_url = url
-        self._settings = settings
-        self._credentials = _get_credentials(credentials)
-        self._update_tokens()
+        self._external_token = _get_external_token(tokens_file)
+        if not self._external_token:
+            self._credentials = _get_credentials(credentials)
+            self._update_tokens()
 
-    def submit_circuit(
+    def __del__(self):
+        try:
+            # try our best to close the auth session, doesn't matter if it fails,
+            # refresh token will be re-issued for the same credentials or eventually expire
+            if not self._external_token:
+                self.close_auth_session()
+        except Exception:  # pylint: disable=broad-except
+            pass
+
+    # pylint: disable=too-many-locals
+    def submit_circuits(
             self,
-            circuit: Circuit,
-            qubit_mapping: Optional[list[SingleQubitMapping]] = None,
-            shots: int = 1
+            circuits: list[Circuit],
+            *,
+            qubit_mapping: Optional[dict[str, str]] = None,
+            settings: Optional[dict[str, Any]] = None,
+            calibration_set_id: Optional[int] = None,
+            shots: int = 1,
     ) -> UUID:
-        """Submits a quantum circuit to be executed on a quantum computer.
+        """Submits a batch of quantum circuits for execution on a quantum computer.
 
         Args:
-            circuit: circuit to be executed
-            qubit_mapping: Mapping of human-readable (logical) qubit names in ``circuit`` to physical qubit names.
-                Can be set to ``None`` if ``circuit`` already uses physical qubit names.
+            circuits: list of circuit to be executed
+            qubit_mapping: Mapping of human-readable (logical) qubit names in to physical qubit names.
+                Can be set to ``None`` if all ``circuits`` already use physical qubit names.
+                Note that the ``qubit_mapping`` is used for all ``circuits``.
+            settings: settings for the quantum computer
+            calibration_set_id: ID of the calibration set to use instead of ``settings``
             shots: number of times ``circuit`` is executed
 
         Returns:
             ID for the created task. This ID is needed to query the status and the execution results.
         """
+        serialized_qubit_mapping: Optional[list[SingleQubitMapping]] = None
+        if qubit_mapping is not None:
+            # check if qubit mapping is injective
+            target_qubits = set(qubit_mapping.values())
+            if not len(target_qubits) == len(qubit_mapping):
+                raise ValueError('Multiple logical qubits map to the same physical qubit.')
 
+            # check if qubit mapping covers all qubits in the circuits
+            for circuit in circuits:
+                circuit_qubits = circuit.all_qubits()
+                diff = circuit_qubits - set(qubit_mapping.keys())
+                if diff:
+                    raise ValueError(f'The qubits {diff} are not found in the provided qubit mapping.')
+
+            # check that all the physical qubit names in qubit_mapping are defined in the settings
+            if settings is not None:
+                physical_qubits = set(settings['subtrees'])  # pylint: disable=unsubscriptable-object
+                diff = target_qubits - physical_qubits
+                if diff:
+                    raise ValueError(f'The physical qubits {diff} in the qubit mapping are not defined in settings.')
+
+            serialized_qubit_mapping = serialize_qubit_mapping(qubit_mapping)
+
+        # ``bearer_token`` can be ``None`` if cocos we're connecting does not use authentication
         bearer_token = self._get_bearer_token()
 
         data = RunRequest(
-            qubit_mapping=qubit_mapping,
-            circuit=circuit,
-            settings=self._settings,
+            qubit_mapping=serialized_qubit_mapping,
+            circuits=circuits,
+            settings=settings,
+            calibration_set_id=calibration_set_id,
             shots=shots
         )
 
@@ -378,18 +567,23 @@ class IQMClient:
             headers['Authorization'] = bearer_token
 
         result = requests.post(
-            join(self._base_url, 'circuit/run'),
-            json=data.dict(),
+            join(self._base_url, 'jobs'),
+            json=data.dict(exclude_none=True),
             headers=headers,
+            timeout=REQUESTS_TIMEOUT
         )
+
+        if result.status_code == 401:
+            raise ClientConfigurationError(f'Authentication failed: {result.text}')
+
         result.raise_for_status()
         return UUID(result.json()['id'])
 
-    def get_run(self, run_id: UUID) -> RunResult:
-        """Query the status of the running task.
+    def get_run(self, job_id: UUID) -> RunResult:
+        """Query the status and results of the running task.
 
         Args:
-            run_id: id of the task
+            job_id: id of the task
 
         Returns:
             result of the run (can be pending)
@@ -400,23 +594,50 @@ class IQMClient:
         """
         bearer_token = self._get_bearer_token()
         result = requests.get(
-            join(self._base_url, 'circuit/run/', str(run_id)),
-            headers=None if not bearer_token else {'Authorization': bearer_token}
+            join(self._base_url, 'jobs/', str(job_id)),
+            headers=None if not bearer_token else {'Authorization': bearer_token},
+            timeout=REQUESTS_TIMEOUT
         )
         result.raise_for_status()
-        result = RunResult.from_dict(result.json())
-        if result.warnings:
-            for warning in result.warnings:
+        run_result = RunResult.from_dict(result.json())
+        if run_result.warnings:
+            for warning in run_result.warnings:
                 warnings.warn(warning)
-        if result.status == RunStatus.FAILED:
-            raise CircuitExecutionError(result.message)
-        return result
+        if run_result.status == Status.FAILED:
+            raise CircuitExecutionError(run_result.message)
+        return run_result
 
-    def wait_for_results(self, run_id: UUID, timeout_secs: float = DEFAULT_TIMEOUT_SECONDS) -> RunResult:
+    def get_run_status(self, job_id: UUID) -> RunStatus:
+        """Query the status of the running task.
+
+        Args:
+            job_id: id of the task
+
+        Returns:
+            status of the run
+
+        Raises:
+            HTTPException: http exceptions
+            CircuitExecutionError: IQM server specific exceptions
+        """
+        bearer_token = self._get_bearer_token()
+        result = requests.get(
+            join(self._base_url, 'jobs/', str(job_id), 'status'),
+            headers=None if not bearer_token else {'Authorization': bearer_token},
+            timeout=REQUESTS_TIMEOUT
+        )
+        result.raise_for_status()
+        run_result = RunStatus.from_dict(result.json())
+        if run_result.warnings:
+            for warning in run_result.warnings:
+                warnings.warn(warning)
+        return run_result
+
+    def wait_for_results(self, job_id: UUID, timeout_secs: float = DEFAULT_TIMEOUT_SECONDS) -> RunResult:
         """Poll results until run is ready, failed, or timed out.
 
         Args:
-            run_id: id of the task to wait
+            job_id: id of the task to wait
             timeout_secs: how long to wait for a response before raising an APITimeoutError
 
         Returns:
@@ -427,30 +648,37 @@ class IQMClient:
         """
         start_time = datetime.now()
         while (datetime.now() - start_time).total_seconds() < timeout_secs:
-            results = self.get_run(run_id)
-            if results.status != RunStatus.PENDING:
+            results = self.get_run(job_id)
+            if results.status != Status.PENDING:
                 return results
             time.sleep(SECONDS_BETWEEN_CALLS)
         raise APITimeoutError(f"The task didn't finish in {timeout_secs} seconds.")
 
-    def close(self) -> bool:
-        """Terminate session with authentication server.
+    def close_auth_session(self) -> bool:
+        """Terminate session with authentication server if there was one created.
 
         Returns:
             True iff session was successfully closed
 
         Raises:
             ClientAuthenticationError: if logout failed
+            ClientAuthenticationError: if asked to close externally managed authentication session
         """
+        # auth session is managed externally, unable to close it here
+        if self._external_token:
+            raise ClientAuthenticationError('Unable to close externally managed auth session')
+
+        # no auth, nothing to close
         if self._credentials is None:
             return False
 
+        # auth session wasn't started, nothing to close
         if not self._credentials.refresh_token:
             return False
 
         url = f'{self._credentials.auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout'
         data = AuthRequest(client_id=AUTH_CLIENT_ID, refresh_token=self._credentials.refresh_token)
-        result = requests.post(url, data=data.dict(exclude_none=True))
+        result = requests.post(url, data=data.dict(exclude_none=True), timeout=REQUESTS_TIMEOUT)
         if result.status_code not in [200, 204]:
             raise ClientAuthenticationError(f'Logout failed, {result.text}')
         self._credentials.access_token = None
@@ -467,6 +695,8 @@ class IQMClient:
             Bearer token, i.e. string containing prefix 'Bearer ' and the access token, or None if access token
             is not available.
         """
+        if self._external_token:
+            return f'Bearer {self._external_token.access_token}'
         if self._credentials is None or not self._credentials.access_token:
             return None
         if _time_left_seconds(self._credentials.access_token) > REFRESH_MARGIN_SECONDS:
@@ -509,7 +739,7 @@ class IQMClient:
             )
 
         url = f'{self._credentials.auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/token'
-        result = requests.post(url, data=data.dict(exclude_none=True))
+        result = requests.post(url, data=data.dict(exclude_none=True), timeout=REQUESTS_TIMEOUT)
         if result.status_code != 200:
             raise ClientAuthenticationError(f'Failed to update tokens, {result.text}')
         tokens = result.json()
