@@ -28,9 +28,11 @@ import requests
 from mockito import expect, mock, unstub, when
 from requests import HTTPError, Response
 
-from iqm_client.iqm_client import (AUTH_CLIENT_ID, AUTH_REALM, AuthRequest,
-                                   GrantType)
+from iqm_client import AUTH_CLIENT_ID, AUTH_REALM, AuthRequest, GrantType
 
+REQUESTS_TIMEOUT = 10
+
+calibration_set_id_value = 24
 existing_run = UUID('3c3fcda3-e860-46bf-92a4-bcc59fa76ce9')
 missing_run = UUID('059e4186-50a3-4e6c-ba1f-37fe6afbdfc2')
 
@@ -50,11 +52,11 @@ def credentials():
 
 
 @pytest.fixture(scope='function')
-def mock_server(base_url):
+def mock_server(base_url, sample_circuit):
     """
     Runs mocking separately for each test
     """
-    generate_server_stubs(base_url)
+    generate_server_stubs(base_url, sample_circuit)
     yield  # running test function
     unstub()
 
@@ -69,6 +71,21 @@ def settings_dict():
         return json.loads(f.read())
 
 
+@pytest.fixture()
+def calibration_set_id():
+    return calibration_set_id_value
+
+
+@pytest.fixture
+def tokens_dict():
+    """
+    Reads and parses tokens file into a dictionary
+    """
+    tokens_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/tokens.json'
+    with open(tokens_path, 'r', encoding='utf-8') as f:
+        return json.loads(f.read())
+
+
 @pytest.fixture
 def sample_circuit():
     """
@@ -91,10 +108,8 @@ def sample_circuit():
                     'Qubit A'
                 ],
                 'args': {
-                    'phase_t': 1.22,
-                    'angle_t': {
-                        'expr': '{{alpha}}/2'
-                    }
+                    'phase_t': 0.7,
+                    'angle_t': 0.25
                 }
             },
             {
@@ -127,28 +142,39 @@ class MockJsonResponse:
             raise HTTPError('')
 
 
-def generate_server_stubs(base_url):
+def generate_server_stubs(base_url, sample_circuit):
     """
     Mocking some calls to the server by mocking 'requests'
     """
-    when(requests).post(f'{base_url}/circuit/run', ...).thenReturn(
+    when(requests).post(f'{base_url}/jobs', ...).thenReturn(
         MockJsonResponse(201, {'id': str(existing_run)})
     )
 
-    when(requests).get(f'{base_url}/circuit/run/{existing_run}', ...).thenReturn(
-        MockJsonResponse(200, {'status': 'pending'})
+    when(requests).get(f'{base_url}/jobs/{existing_run}', ...).thenReturn(
+        MockJsonResponse(200, {'status': 'pending', 'metadata': {'shots': 42, 'circuits': [sample_circuit]}})
     ).thenReturn(
         MockJsonResponse(
             200,
-            {'status': 'ready', 'measurements': {'result': [[1, 0, 1, 1], [1, 0, 0, 1], [1, 0, 1, 1], [1, 0, 1, 1]]}}
+            {
+                'status': 'ready',
+                'measurements': [{'result': [[1, 0, 1, 1], [1, 0, 0, 1], [1, 0, 1, 1], [1, 0, 1, 1]]}],
+                'metadata': {'shots': 42, 'circuits': [sample_circuit], 'calibration_set_id': calibration_set_id_value}
+            }
         )
     )
 
+    when(requests).get(f'{base_url}/jobs/{existing_run}/status', ...).thenReturn(
+        MockJsonResponse(200, {'status': 'pending'})
+    ).thenReturn(
+        MockJsonResponse(200, {'status': 'ready'})
+    )
+
     # 'run was not created' response
     no_run_response = Response()
     no_run_response.status_code = 404
     no_run_response.reason = 'Run not found'
-    when(requests).get(f'{base_url}/circuit/run/{missing_run}', ...).thenReturn(no_run_response)
+    when(requests).get(f'{base_url}/jobs/{missing_run}', ...).thenReturn(no_run_response)
+    when(requests).get(f'{base_url}/jobs/{missing_run}/status', ...).thenReturn(no_run_response)
 
 
 def prepare_tokens(
@@ -190,7 +216,8 @@ def prepare_tokens(
     }
     when(requests).post(
         f'{credentials["auth_server_url"]}/realms/{AUTH_REALM}/protocol/openid-connect/token',
-        data=request_data.dict(exclude_none=True)
+        data=request_data.dict(exclude_none=True),
+        timeout=REQUESTS_TIMEOUT
     ).thenReturn(MockJsonResponse(status_code, tokens))
 
     return tokens
@@ -226,8 +253,8 @@ def expect_status_request(url: str, access_token: Optional[str], times: int = 1)
     """
     job_id = uuid4()
     headers = None if access_token is None else {'Authorization': f'Bearer {access_token}'}
-    expect(requests, times=times).get(f'{url}/circuit/run/{job_id}', headers=headers).thenReturn(
-        MockJsonResponse(200, {'status': 'pending'})
+    expect(requests, times=times).get(f'{url}/jobs/{job_id}', headers=headers, timeout=REQUESTS_TIMEOUT).thenReturn(
+        MockJsonResponse(200, {'status': 'pending', 'metadata': {'shots': 42, 'circuits': []}})
     )
     return job_id
 
@@ -242,7 +269,8 @@ def expect_logout(auth_server_url: str, refresh_token: str):
     request_data = AuthRequest(client_id=AUTH_CLIENT_ID, refresh_token=refresh_token)
     expect(requests, times=1).post(
         f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout',
-        data=request_data.dict(exclude_none=True)
+        data=request_data.dict(exclude_none=True),
+        timeout=REQUESTS_TIMEOUT
     ).thenReturn(
         mock({'status_code': 204, 'text': '{}'})
     )
